package mvd.jester.model;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import com.google.common.collect.Iterables;
import com.google.common.collect.Sets;
import org.jgrapht.Graphs;
import org.jgrapht.experimental.dag.DirectedAcyclicGraph;
import org.jgrapht.graph.DefaultEdge;
import mvd.jester.utils.BinaryDecompositionTree;
import mvd.jester.utils.BinaryDecompositionTree.Node;
import mvd.jester.utils.BinaryDecompositionTree.NodeType;

public class DagTask implements Task {

    private DirectedAcyclicGraph<Job, DefaultEdge> jobDag;
    private final Set<Segment> workloadDistribution;
    private final long workload;
    private final long criticalPath;
    private final long period;
    private final long deadline;
    private final long numberOfThreads;

    public DagTask(DirectedAcyclicGraph<Job, DefaultEdge> jobDag, long period,
            long numberOfThreads) {
        this.jobDag = jobDag;
        this.period = period;
        this.deadline = period;
        this.numberOfThreads = numberOfThreads;
        this.workload = DagUtils.calculateWorkload(this.jobDag);
        this.criticalPath = DagUtils.calculateCriticalPath(this.jobDag);
        this.workloadDistribution =
                DagUtils.calculateWorkloadDistribution(this.jobDag, this.criticalPath);
    }

    public double getUtilization() {
        return (double) workload / period;
    }

    /**
     * @return the deadline
     */
    public long getDeadline() {
        return deadline;
    }

    /**
     * @return the jobDag
     */
    public DirectedAcyclicGraph<Job, DefaultEdge> getJobDag() {
        return jobDag;
    }

    /**
     * @param jobDag the jobDag to set
     */
    public void setJobDag(DirectedAcyclicGraph<Job, DefaultEdge> jobDag) {
        this.jobDag = jobDag;
    }

    /**
     * @return the period
     */
    public long getPeriod() {
        return period;
    }

    /**
     * @return the workload
     */
    public long getWorkload() {
        return workload;
    }

    /**
     * @return the criticalPath
     */
    public long getCriticalPath() {
        return criticalPath;
    }


    /**
     * @return the workloadDistribution
     */
    public Set<Segment> getWorkloadDistribution() {
        return workloadDistribution;
    }

    @Override
    public long getMaximumParallelism() {
        long max = 0;
        for (Segment s : workloadDistribution) {
            if (max < s.getNumberOfJobs()) {
                max = s.getNumberOfJobs();
            }
        }
        return max;
    }


    @Override
    public long getNumberOfThreads() {
        return numberOfThreads;
    }

    public static class DagUtils {
        public static long calculateWorkload(DirectedAcyclicGraph<Job, DefaultEdge> jobDag) {
            long workload = 0;

            for (Job job : jobDag) {
                workload += job.getWcet();
            }
            return workload;
        }

        public static long calculateCriticalPath(DirectedAcyclicGraph<Job, DefaultEdge> jobDag) {
            long criticalPath = 0;
            // BreadthFirstIterator<Job, DefaultEdge> breadthFirstIterator =
            // new BreadthFirstIterator<>(jobDag);
            // while (breadthFirstIterator.hasNext()) {
            // Job job = breadthFirstIterator.next();
            for (Job job : jobDag) {
                Set<DefaultEdge> edges = jobDag.incomingEdgesOf(job);
                long longestRelativeCompletionTime = 0;
                for (DefaultEdge e : edges) {
                    Job source = jobDag.getEdgeSource(e);
                    longestRelativeCompletionTime =
                            longestRelativeCompletionTime >= source.getRelativeCompletionTime()
                                    ? longestRelativeCompletionTime
                                    : source.getRelativeCompletionTime();
                }

                job.setRelativeCompletionTime(longestRelativeCompletionTime + job.getWcet());
                criticalPath = job.getRelativeCompletionTime();
            }

            return criticalPath;
        }

        public static LinkedHashSet<Segment> calculateWorkloadDistribution(
                DirectedAcyclicGraph<Job, DefaultEdge> jobDag, long criticalPath) {
            LinkedHashSet<Segment> segments = new LinkedHashSet<>();
            long segmentDuration = 0;
            long segmentHeight = 1;
            for (long t = 0; t < criticalPath; ++t) {
                long currentHeight = 0;
                for (Job j : jobDag) {
                    if (t >= j.getRelativeCompletionTime() - j.getWcet()
                            && t < j.getRelativeCompletionTime()) {
                        currentHeight++;
                    }
                }
                if (currentHeight == segmentHeight) {
                    segmentDuration++;
                } else {
                    segments.add(new Segment(segmentDuration, segmentHeight));
                    segmentDuration = 1;
                    segmentHeight = currentHeight;
                }
            }
            segments.add(new Segment(segmentDuration, segmentHeight));

            return segments;
        }

        public static DirectedAcyclicGraph<Job, DefaultEdge> createNFJGraph(
                DirectedAcyclicGraph<Job, DefaultEdge> jobDag) {
            LinkedList<Job> joinNodes = new LinkedList<>();
            LinkedList<Job> forkNodes = new LinkedList<>();

            collectForksAndJoins(jobDag, forkNodes, joinNodes);
            return createNFJGraph(jobDag, forkNodes, joinNodes);
        }

        public static void collectForksAndJoins(DirectedAcyclicGraph<Job, DefaultEdge> jobDag,
                LinkedList<Job> forkNodes, LinkedList<Job> joinNodes) {
            for (Job j : jobDag) {
                if (jobDag.inDegreeOf(j) > 1) {
                    joinNodes.add(j);
                }
                if (jobDag.outDegreeOf(j) > 1) {
                    forkNodes.add(j);
                }
            }

            for (Job j : forkNodes) {
                if (jobDag.outDegreeOf(j) <= 1) {
                    int test = 3;
                }
            }
        }

        public static DirectedAcyclicGraph<Job, DefaultEdge> createNFJGraph(
                DirectedAcyclicGraph<Job, DefaultEdge> jobDag, LinkedList<Job> forkNodes,
                LinkedList<Job> joinNodes) {
            DirectedAcyclicGraph<Job, DefaultEdge> modifiedJobDag =
                    new DirectedAcyclicGraph<>(DefaultEdge.class);
            Graphs.addGraph(modifiedJobDag, jobDag);

            if (!(joinNodes.size() > 1)) {
                return modifiedJobDag;
            }

            final Job sink = joinNodes.getLast();

            for (Job j : joinNodes) {
                Set<DefaultEdge> edgeSet = new HashSet<>(modifiedJobDag.incomingEdgesOf(j));
                for (DefaultEdge e : edgeSet) {
                    Job predecessor = modifiedJobDag.getEdgeSource(e);
                    boolean satisfiesProposition =
                            DagUtils.checkForFork(modifiedJobDag, j, forkNodes, predecessor);
                    if (!satisfiesProposition) {
                        modifiedJobDag.removeEdge(e);
                        if (modifiedJobDag.outgoingEdgesOf(predecessor).isEmpty()) {
                            try {
                                modifiedJobDag.addDagEdge(predecessor, sink);
                            } catch (Exception ex) {
                            }
                        }
                    }

                    if (modifiedJobDag.inDegreeOf(j) == 1) {
                        break;
                    }
                }
            }

            return modifiedJobDag;
        }

        public static BinaryDecompositionTree<Job> createDecompositionTree(
                DirectedAcyclicGraph<Job, DefaultEdge> jobDag) {
            LinkedList<Job> joinNodes = new LinkedList<>();
            LinkedList<Job> forkNodes = new LinkedList<>();

            collectForksAndJoins(jobDag, forkNodes, joinNodes);
            return createDecompositionTree(jobDag, forkNodes, joinNodes);
        }

        public static BinaryDecompositionTree<Job> createDecompositionTree(
                DirectedAcyclicGraph<Job, DefaultEdge> jobDag, LinkedList<Job> forkNodes,
                LinkedList<Job> joinNodes) {
            BinaryDecompositionTree<Job> tree = new BinaryDecompositionTree<>();
            Optional<Job> source = Optional.empty();
            Optional<Job> sink = Optional.empty();
            if (forkNodes.size() > 0) {
                source = Optional.of(forkNodes.getFirst());
                sink = Optional.of(joinNodes.getLast());
            } else {
                boolean firstRun = true;
                for (Job j : jobDag) {
                    if (firstRun) {
                        firstRun = false;
                        source = Optional.of(j);
                    }
                    sink = Optional.of(j);
                }
            }
            if (source.isPresent() && sink.isPresent()) {
                Job current = source.get();

                DagUtils.constructTree(jobDag, null, current, tree, forkNodes, joinNodes,
                        sink.get());
            }

            return tree;
        }

        private static Node<Job> constructTree(DirectedAcyclicGraph<Job, DefaultEdge> jobDag,
                Node<Job> parentNode, Job currentJob, BinaryDecompositionTree<Job> tree,
                LinkedList<Job> forkVertices, LinkedList<Job> joinVertices, Job sinkJob) {
            if (forkVertices.contains(currentJob)) {
                Job forkJob = currentJob;
                Job joinJob = findJoin(jobDag, forkJob, sinkJob, joinVertices);

                Node<Job> seqForkNode = new Node<Job>(parentNode, NodeType.SEQ);
                seqForkNode.setLeftNode(new Node<Job>(seqForkNode, forkJob));
                if (tree.isEmpty()) {
                    tree.setRoot(seqForkNode);
                }
                final Node<Job> parContinuationNode;
                if (!tree.contains(joinJob)) {
                    Node<Job> seqJoinNode =
                            seqForkNode.setRightNode(new Node<Job>(seqForkNode, NodeType.SEQ));

                    Node<Job> subTreeOfJoin = constructTree(jobDag, seqJoinNode, joinJob, tree,
                            forkVertices, joinVertices, sinkJob);
                    seqJoinNode.setRightNode(subTreeOfJoin);

                    parContinuationNode =
                            seqJoinNode.setLeftNode(new Node<Job>(seqJoinNode, NodeType.PAR));
                } else {
                    parContinuationNode =
                            seqForkNode.setRightNode(new Node<Job>(seqForkNode, NodeType.PAR));
                }


                Node<Job> successorParent = parContinuationNode;
                List<Job> successors = Graphs.successorListOf(jobDag, currentJob);
                if (successors.size() == 1) {
                    int test = 3;
                }
                // create leftSide of joinNode
                for (int i = 0; i < successors.size(); ++i) {
                    Job succ = successors.get(i);
                    Node<Job> thisNode = constructTree(jobDag, successorParent, succ, tree,
                            forkVertices, joinVertices, sinkJob);
                    if (i == successors.size() - 1) {
                        successorParent.setRightNode(thisNode);
                    } else if (i == successors.size() - 2) {
                        successorParent.setLeftNode(thisNode);
                    } else {
                        successorParent.setLeftNode(thisNode);
                        successorParent.setRightNode(new Node<>(successorParent, NodeType.PAR));
                        successorParent = successorParent.getRightNode();
                    }
                }

                return seqForkNode;
            } else {
                List<Job> successorList = Graphs.successorListOf(jobDag, currentJob);
                if (successorList.size() > 0) {
                    Job successor = successorList.get(0);
                    if (!tree.contains(successor)) {
                        Node<Job> seqJobNode = new Node<Job>(parentNode, NodeType.SEQ);
                        if (tree.isEmpty()) {
                            tree.setRoot(seqJobNode);
                        }
                        seqJobNode.setLeftNode(new Node<Job>(seqJobNode, currentJob));
                        Node<Job> contNode = constructTree(jobDag, seqJobNode, successor, tree,
                                forkVertices, joinVertices, sinkJob);
                        seqJobNode.setRightNode(contNode);
                        return seqJobNode;
                    }
                }
                Node<Job> jobNode = new Node<Job>(parentNode, currentJob);
                if (tree.isEmpty()) {
                    tree.setRoot(jobNode);
                }
                return jobNode;
            }
        }

        public static boolean checkNFJProperty(DirectedAcyclicGraph<Job, DefaultEdge> jobDag) {
            LinkedList<Job> joinNodes = new LinkedList<>();
            LinkedList<Job> forkNodes = new LinkedList<>();

            collectForksAndJoins(jobDag, forkNodes, joinNodes);

            if (!(joinNodes.size() > 1)) {
                return true;
            }

            nextJoin: for (Job j : joinNodes) {
                for (Job f : forkNodes) {
                    Set<Job> ancestorsOfJ = jobDag.getAncestors(jobDag, j);
                    Set<Job> ancesotorsOfF = jobDag.getAncestors(jobDag, f);
                    ancesotorsOfF.add(f);
                    Set<Job> inbetweenJobs = new HashSet<>(ancestorsOfJ);
                    inbetweenJobs.removeAll(ancesotorsOfF);
                    if (inbetweenJobs.isEmpty()) {
                        continue;
                    }

                    for (Job a : inbetweenJobs) {
                        List<Job> neighboursOfA = Graphs.neighborListOf(jobDag, a);
                        for (Job b : neighboursOfA) {
                            if ((jobDag.getAncestors(jobDag, j).contains(b) || b == j)
                                    && (jobDag.getDescendants(jobDag, f).contains(b) || b == f)) {
                                continue nextJoin;
                            }
                        }
                    }
                }
                return false;
            }
            return true;
        }

        private static Job findJoin(DirectedAcyclicGraph<Job, DefaultEdge> jobDag, Job forkNode,
                Job sink, List<Job> joinNodes) {
            for (Job j : joinNodes) {
                Set<Job> ancestorsOfJ = jobDag.getAncestors(jobDag, j);
                Set<Job> ancesotorsOfFork = jobDag.getAncestors(jobDag, forkNode);
                ancesotorsOfFork.add(forkNode);
                Set<Job> inbetweenJobs = new HashSet<>(ancestorsOfJ);
                inbetweenJobs.removeAll(ancesotorsOfFork);
                if (inbetweenJobs.isEmpty()) {
                    continue;
                }

                List<Job> successorOfFork = Graphs.successorListOf(jobDag, forkNode);

                if (inbetweenJobs.containsAll(successorOfFork)) {
                    return j;
                }
            }

            return sink;
        }

        private static boolean checkForFork(DirectedAcyclicGraph<Job, DefaultEdge> jobDag,
                Job joinNode, List<Job> forkNodes, Job job) {
            List<Job> pred = Graphs.predecessorListOf(jobDag, job);

            for (Job p : pred) {
                if (forkNodes.contains(p)) {
                    for (Job successor : Graphs.successorListOf(jobDag, p)) {
                        if (jobDag.getAncestors(jobDag, joinNode).contains(successor)) {
                            return false;
                        }
                    }
                } else {
                    return checkForFork(jobDag, joinNode, forkNodes, p);
                }
            }

            return true;
        }


        public static DirectedAcyclicGraph<Job, DefaultEdge> createNFJfromDecompositionTree(
                BinaryDecompositionTree<Job> tree) {
            DirectedAcyclicGraph<Job, DefaultEdge> jobDag =
                    new DirectedAcyclicGraph<>(DefaultEdge.class);

            if (tree.getRootNode() != null) {
                traverseNodes(jobDag, tree.getRootNode());
            }

            return jobDag;

        }

        private static GraphEndPoints traverseNodes(DirectedAcyclicGraph<Job, DefaultEdge> jobDag,
                Node<Job> node) {
            if (node.getNodeType() != NodeType.LEAF
                    && (node.getLeftNode() == null || node.getRightNode() == null)) {
                int test = 3;
            }
            switch (node.getNodeType()) {
                case LEAF: {
                    Job j = node.getObject();
                    jobDag.addVertex(j);
                    Set<Job> endPoints = new HashSet<Job>(Arrays.asList(j));
                    return new GraphEndPoints(endPoints, endPoints);
                }
                case SEQ: {
                    GraphEndPoints leftEndPoints = traverseNodes(jobDag, node.getLeftNode());
                    GraphEndPoints rightEndPoints = traverseNodes(jobDag, node.getRightNode());
                    for (Job l : leftEndPoints.right) {
                        for (Job r : rightEndPoints.left) {
                            if (r == l) {
                                continue;
                            }
                            try {
                                jobDag.addDagEdge(l, r);
                            } catch (Exception e) {
                                int test = 3;
                            }
                        }
                    }

                    return new GraphEndPoints(leftEndPoints.left, rightEndPoints.right);
                }
                case PAR: {
                    GraphEndPoints leftEndPoints = traverseNodes(jobDag, node.getLeftNode());
                    GraphEndPoints rightEndPoints = traverseNodes(jobDag, node.getRightNode());
                    Set<Job> leftEndPointJobs = Sets
                            .newHashSet(Iterables.concat(leftEndPoints.left, rightEndPoints.left));
                    Set<Job> rightEndPointJobs = Sets.newHashSet(
                            Iterables.concat(leftEndPoints.right, rightEndPoints.right));
                    return new GraphEndPoints(leftEndPointJobs, rightEndPointJobs);
                }
                default:
                    break;
            }
            return new GraphEndPoints(new HashSet<>(), new HashSet<>());
        }

        public static class GraphEndPoints {
            public Set<Job> left;
            public Set<Job> right;

            public GraphEndPoints(Set<Job> left, Set<Job> right) {
                this.left = left;
                this.right = right;
            }
        }

        private enum TraversalOrder {
            LEFT, RIGHT
        }

    }
}
