package mvd.jester.tests;

import java.math.RoundingMode;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import javax.print.attribute.standard.JobName;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import com.google.common.math.LongMath;
import org.jgrapht.experimental.dag.DirectedAcyclicGraph;
import org.jgrapht.graph.DefaultEdge;
import mvd.jester.info.SchedulingInfo;
import mvd.jester.info.TerminationInfo;
import mvd.jester.info.TerminationInfo.Level;
import mvd.jester.model.DagTask;
import mvd.jester.model.Job;
import mvd.jester.model.Segment;
import mvd.jester.model.SortedTaskSet;
import mvd.jester.model.Task;
import mvd.jester.model.TreeJob;
import mvd.jester.model.DagTask.DagUtils;
import mvd.jester.priority.PriorityManager;
import mvd.jester.priority.RateMonotonic;
import mvd.jester.utils.BinaryDecompositionTree;
import mvd.jester.utils.BinaryDecompositionTree.Node;
import mvd.jester.utils.BinaryDecompositionTree.NodeType;

public class FonsecaNelis extends AbstractTest<DagTask> {

    private final Map<Task, TerminationInfo> responseTimes;
    private final PriorityManager priorityManager;
    private final Map<Task, Set<Segment>> sortedSegments;

    public FonsecaNelis(long numberOfProcessors) {
        super(numberOfProcessors);
        this.responseTimes = new HashMap<>();
        this.priorityManager = new RateMonotonic();
        this.sortedSegments = new HashMap<>();
    }

    @Override
    public PriorityManager getPriorityManager() {
        return priorityManager;
    }

    @Override
    public SchedulingInfo runSchedulabilityCheck(SortedTaskSet<DagTask> tasks) {
        sortedSegments.clear();
        responseTimes.clear();
        createNFJandDecompositionTree(tasks);
        for (DagTask t : tasks) {
            long responseTime = calculateResponseTime(tasks, t);
            responseTimes.put(t, new TerminationInfo(t.getDeadline(), responseTime, Level.HIGH));
        }

        return new SchedulingInfo(new HashSet<>(responseTimes.values()),
                tasks.getParallelTaskRatio(), tasks.getUtilization());
    }

    private void createNFJandDecompositionTree(SortedTaskSet<DagTask> tasks) {
        for (DagTask t : tasks) {
            DirectedAcyclicGraph<Job, DefaultEdge> jobDag = t.getJobDag();

            DirectedAcyclicGraph<Job, DefaultEdge> nfjJobDag = DagUtils.createNFJGraph(jobDag);
            BinaryDecompositionTree<Job> tree = DagUtils.createDecompositionTree(nfjJobDag);

            sortedSegments.put(t, constructCarryOutDistribution(nfjJobDag, tree));
        }
    }

    private Set<Segment> constructCarryOutDistribution(
            DirectedAcyclicGraph<Job, DefaultEdge> nfjDag, BinaryDecompositionTree<Job> tree) {
        Set<Segment> carryOutWorkload = new LinkedHashSet<>();
        BinaryDecompositionTree<TreeJob> modifiedTree = transformTree(tree);

        boolean isEmpty = false;
        do {
            Set<TreeJob> parallelJobs = getMaximumParallelism(modifiedTree.getRootNode());
            Optional<TreeJob> min =
                    parallelJobs.stream().min((p1, p2) -> Long.compare(p1.getWcet(), p2.getWcet()));
            if (min.isPresent()) {
                long width = min.get().getWcet();
                carryOutWorkload.add(new Segment(width, parallelJobs.size()));
                for (TreeJob p : parallelJobs) {
                    p.setWcet(p.getWcet() - width);
                }
            } else {
                break;
            }
            isEmpty = parallelJobs.isEmpty();
        } while (!isEmpty);

        return carryOutWorkload;
    }


    private BinaryDecompositionTree<TreeJob> transformTree(BinaryDecompositionTree<Job> tree) {
        BinaryDecompositionTree<TreeJob> modifiedTree = new BinaryDecompositionTree<>();
        Node<TreeJob> root = transformNode(null, tree.getRootNode());
        modifiedTree.setRoot(root);
        return modifiedTree;
    }

    private Node<TreeJob> transformNode(Node<TreeJob> parent, Node<Job> node) {
        if (node.getNodeType() != NodeType.LEAF
                && (node.getLeftNode() == null || node.getRightNode() == null)) {
            int test = 3;
        }
        if (node.getNodeType().equals(NodeType.LEAF)) {
            return new Node<TreeJob>(parent, new TreeJob(node.getObject()));
        } else {
            Node<TreeJob> modifiedNode = new Node<TreeJob>(null, node.getNodeType());
            modifiedNode.setLeftNode(transformNode(modifiedNode, node.getLeftNode()));
            modifiedNode.setRightNode(transformNode(modifiedNode, node.getRightNode()));
            return modifiedNode;
        }
    }

    private Set<TreeJob> getMaximumParallelism(Node<TreeJob> node) {
        NodeType nodeType = node.getNodeType();
        if (nodeType.equals(NodeType.PAR)) {
            Set<TreeJob> left = getMaximumParallelism(node.getLeftNode());
            Set<TreeJob> right = getMaximumParallelism(node.getRightNode());
            return Sets.union(left, right);
        } else if (nodeType.equals(NodeType.SEQ)) {
            Set<TreeJob> left = getMaximumParallelism(node.getLeftNode());
            Set<TreeJob> right = getMaximumParallelism(node.getRightNode());
            if (left.size() >= right.size()) {
                return left;
            } else {
                return right;
            }
        } else {
            if (node.getObject().getWcet() > 0) {
                return new HashSet<>(Arrays.asList(node.getObject()));
            } else {
                return new HashSet<>();
            }
        }
    }

    private long calculateResponseTime(SortedTaskSet<DagTask> tasks, DagTask task) {
        long criticalPath = task.getCriticalPath();
        long responseTime = criticalPath;
        long previousResponseTime = 0;

        do {
            previousResponseTime = responseTime;
            double taskInterference = 0;
            for (DagTask t : tasks) {
                if (t.getPeriod() < task.getPeriod()) {
                    taskInterference += getTaskInterference(t, responseTime);
                }
            }

            taskInterference /= numberOfProcessors;

            double selfInterference = getSelfInterference(task);

            long totalInterference = (long) Math.floor(taskInterference + selfInterference);
            responseTime = criticalPath + totalInterference;

        } while (responseTime != previousResponseTime);

        return responseTime;
    }

    private double getTaskInterference(DagTask task, long interval) {
        long period = task.getPeriod();
        long criticalPath = task.getCriticalPath();
        long numberOfIntervals =
                LongMath.divide(interval - criticalPath, period, RoundingMode.FLOOR);
        long carryInAndOutInterval = interval - Math.max(0, numberOfIntervals) * period;
        long numberOfBodyJobs =
                LongMath.divide(interval - carryInAndOutInterval, period, RoundingMode.FLOOR);

        long bodyWorkload = Math.max(0, numberOfBodyJobs) * task.getWorkload();


        long carryInAndOutWorkload = getCarryInAndOutWorkload(task, task.getWorkloadDistribution(),
                sortedSegments.get(task), carryInAndOutInterval);

        return carryInAndOutWorkload + bodyWorkload;
    }

    private long getCarryInAndOutWorkload(DagTask task, Set<Segment> carryInDistribution,
            Set<Segment> carryOutDistribution, long carryInAndOutPeriod) {
        long workload = getCarryOutWorkload(task, carryOutDistribution, carryInAndOutPeriod);
        long carryInPeriod = task.getPeriod() - responseTimes.get(task).getResponseTime();
        long carryOutPeriod = 0;
        List<Segment> carryInList = Lists.newArrayList(carryInDistribution);
        Collections.reverse(carryInList);

        for (Segment s : carryInList) {
            carryInPeriod += s.getJobWcet();
            carryOutPeriod = carryInAndOutPeriod - carryInPeriod;
            long carryInWorkload = getCarryInWorkload(task, carryInDistribution, carryInPeriod);
            long carryOutWorkload = getCarryOutWorkload(task, carryOutDistribution, carryOutPeriod);
            workload = Math.max(workload, carryInWorkload + carryOutWorkload);
        }

        workload = Math.max(workload,
                getCarryInWorkload(task, carryInDistribution, carryInAndOutPeriod));
        carryOutPeriod = 0;

        for (Segment s : carryOutDistribution) {
            carryOutPeriod += s.getJobWcet();
            carryInPeriod = carryInAndOutPeriod - carryOutPeriod;
            long carryInWorkload = getCarryInWorkload(task, carryInDistribution, carryInPeriod);
            long carryOutWorkload = getCarryOutWorkload(task, carryOutDistribution, carryOutPeriod);
            workload = Math.max(workload, carryInWorkload + carryOutWorkload);
        }

        return workload;
    }

    private long getCarryOutWorkload(DagTask task, Set<Segment> carryOutDistribution,
            long carryOutPeriod) {
        long workload = 0;
        long period = task.getPeriod();
        long responseTime = responseTimes.get(task).getResponseTime();
        List<Segment> distributionList = Lists.newArrayList(carryOutDistribution);

        for (int i = 0; i < distributionList.size(); ++i) {
            Segment s = distributionList.get(i);
            long weightOfPreviousSegments = 0;
            for (int j = 0; j < i; ++j) {
                weightOfPreviousSegments += distributionList.get(j).getJobWcet();
            }
            long width = carryOutPeriod - weightOfPreviousSegments;

            workload += Math.max(Math.min(width, s.getJobWcet()), 0) * s.getNumberOfJobs();
        }

        long improvedWorkload =
                Math.max(carryOutPeriod - (period - responseTime), 0) * numberOfProcessors;

        return Math.min(improvedWorkload, workload);
    }

    private long getCarryInWorkload(DagTask task, Set<Segment> carryInDistribution,
            long carryInPeriod) {
        long workload = 0;
        long period = task.getPeriod();
        long responseTime = responseTimes.get(task).getResponseTime();
        List<Segment> distributionList = Lists.newArrayList(carryInDistribution);

        for (int i = 0; i < carryInDistribution.size(); ++i) {
            Segment s = distributionList.get(i);
            long weightOfRemainingSegments = 0;
            for (int j = i + 1; j < carryInDistribution.size(); ++j) {
                weightOfRemainingSegments += distributionList.get(j).getJobWcet();
            }
            long width = carryInPeriod - period + responseTime - weightOfRemainingSegments;

            workload += Math.max(Math.min(width, s.getJobWcet()), 0) * s.getNumberOfJobs();
        }

        long improvedWorkload =
                Math.max(carryInPeriod - (period - responseTime), 0) * numberOfProcessors;

        return Math.min(improvedWorkload, workload);
    }

    private double getSelfInterference(DagTask task) {
        long criticalPath = task.getCriticalPath();
        long workload = task.getWorkload();

        return (double) (workload - criticalPath) / numberOfProcessors;
    }

    @Override
    public String getName() {
        return "FonsecaNelis";
    }


}
